#include "p30F6014A.h"
#include "stdio.h"
#include "string.h"
#include "math.h"
#include <time.h>
#include <stdlib.h>

#include <motor_led/e_init_port.h>
#include <motor_led/advance_one_timer/e_led.h>
#include <motor_led/advance_one_timer/e_motors.h>
#include <motor_led/advance_one_timer/e_agenda.h>
#include <uart/e_uart_char.h>
#include <a_d/advance_ad_scan/e_ad_conv.h>
#include <a_d/advance_ad_scan/e_prox.h>
#include <a_d/advance_ad_scan/e_acc.h>
#include <a_d/advance_ad_scan/e_micro.h>
#include <bluetooth/e_bluetooth.h>
#include <motor_led/e_epuck_ports.h>
#include <acc_gyro/e_lsm330.h>
#include <utility/utility.h>
#include "memory.h"

int main() {
    
    char buffer[BUFFER_SIZE];
    char c;
    int i = 0;
    signed int tempX = 0, tempY = 0, tempZ = 0;
    double speed = 0;
    int gyroYOffset = 0, accYOffset = 0;
    long int gyroYOffsetSum = 0, accYOffsetSum = 0;
    unsigned int offsetCount = 0;
    double pController = 50.0, dController = 75.0, iController = 10.0, iControllerSumMax = 50.0;
    double kGyro = 0.95;
    double kGyroDeltaAngleFactor = 100.0;
    double accY, accZ, accAngle = 0.0;
    double gyroDeltaAngle = 0.0, gyroDps = 0.0;
    double currAngle = 0.0, prevAngle = 0.0, angleZeroThr = 0.0;
    double error = 0, prevError = 0, errorSum = 0;
    unsigned char updateOffset = 0;
    double maxSpeed = 1000.0;
    
    

    //system initialization
    e_init_port();    // configure port pins
    e_start_agendas_processing();
    e_init_motors();
    e_init_uart1();   // initialize UART to 115200 Kbaud
    e_init_uart2(BAUD115200);   // initialize UART to 115200 Kbaud
    e_init_ad_scan(ALL_ADC);
    
    //Reset if Power on (some problem for few robots)
    if (RCONbits.POR) {
        RCONbits.POR=0;
        __asm__ volatile ("reset");
    }

    // Wait the user confirmation (any char) before starting, this is to let the user place the robot and start when
    // the robot is well aligned.
    while(e_getchar_uart1(&c) == 0);
    
    // Calibrate accelerometer and gyroscope.
    e_set_led(8, 1);
    for(i=0; i<128; i++) {
        accYOffsetSum += e_get_acc_filtered(1,1);
        gyroYOffsetSum += getYAxisGyro();
    }
    accYOffset = accYOffsetSum>>7;
    gyroYOffset = gyroYOffsetSum>>7;
    accYOffsetSum = 0;
    gyroYOffsetSum = 0;
    e_set_led(8, 0);

    resetTime();
    
    while (1) {
        LED0 = 1;   // This led is used to measure the controller frequency with an oscilloscope.
        
        // The user can change dinamically the controller parameters through Bluetooth:
        // q, w => increment, decrement P by 0.5
        // a, s => increment, decrement D by 0.5
        // y, x => increment, decrement I by 0.5
        // d, f => increment, decrement "angleZeroThr" by 0.5. This value set the resulting angle to zero
        //         when lower than this threshold and moreover reset the "errorSum"
        // c, v => disable, enable the update of the accelerometer and gyro offsets
        // t, z => increment, decrement by 50.0 the maximum speed given by the controller
        // g, h => increment, decrement by 0.01 the weight of the gyro in the final angle, the weight for the accelerometer will be (1-k).
        // b, n => increment, decrement by 0.5 the factor used to multiply the "gyroDeltaAngle"...from field test this factor gives better results.
        if(e_getchar_uart1(&c) == 1) {
            if(c == 'q') {
                pController += 0.5;
                sprintf(buffer, "%+5.2f\r\n", pController);
                e_send_uart1_char(buffer, strlen(buffer));
            }
            if(c == 'w') {
                pController -= 0.5;
                sprintf(buffer, "%+5.2f\r\n", pController);
                e_send_uart1_char(buffer, strlen(buffer));
            }
            if(c == 'a') {
                dController += 0.5;
                sprintf(buffer, "%+5.2f\r\n", dController);
                e_send_uart1_char(buffer, strlen(buffer));
            }
            if(c == 's') {
                dController -= 0.5;
                sprintf(buffer, "%+5.2f\r\n", dController);
                e_send_uart1_char(buffer, strlen(buffer));
            }
            if(c == 'y') {
                iController += 0.5;
                sprintf(buffer, "%+5.2f\r\n", iController);
                e_send_uart1_char(buffer, strlen(buffer));
            }
            if(c == 'x') {
                iController -= 0.5;
                sprintf(buffer, "%+5.2f\r\n", iController);
                e_send_uart1_char(buffer, strlen(buffer));
            }
            if(c == 'd') {
                angleZeroThr += 0.5;
                sprintf(buffer, "%+5.2f\r\n", angleZeroThr);
                e_send_uart1_char(buffer, strlen(buffer));
            }
            if(c == 'f') {
                angleZeroThr -= 0.5;
                sprintf(buffer, "%+5.2f\r\n", angleZeroThr);
                e_send_uart1_char(buffer, strlen(buffer));
            }
            if(c == 'c') {
                updateOffset = 0;
                sprintf(buffer, "%d\r\n", updateOffset);
                e_send_uart1_char(buffer, strlen(buffer));
            }
            if(c == 'v') {
                updateOffset = 1;
                sprintf(buffer, "%d\r\n", updateOffset);
                e_send_uart1_char(buffer, strlen(buffer));
            }
            if(c == 't') {
                maxSpeed += 50;
                if(maxSpeed > 1000.0) {
                    maxSpeed = 1000.0;
                }
                sprintf(buffer, "%d\r\n", (int)maxSpeed);
                e_send_uart1_char(buffer, strlen(buffer));
            }
            if(c == 'z') {
                maxSpeed -= 50;
                if(maxSpeed < 100.0) {
                    maxSpeed = 100.0;
                }
                sprintf(buffer, "%d\r\n", (int)maxSpeed);
                e_send_uart1_char(buffer, strlen(buffer));
            }
            if(c == 'g') {
                kGyro += 0.01;
                sprintf(buffer, "%+5.2f\r\n", kGyro);
                e_send_uart1_char(buffer, strlen(buffer));
            }
            if(c == 'h') {
                kGyro -= 0.01;
                sprintf(buffer, "%+5.2f\r\n", kGyro);
                e_send_uart1_char(buffer, strlen(buffer));
            }
            if(c == 'b') {
                kGyroDeltaAngleFactor += 0.5;
                sprintf(buffer, "%+5.2f\r\n", kGyroDeltaAngleFactor);
                e_send_uart1_char(buffer, strlen(buffer));
            }
            if(c == 'n') {
                kGyroDeltaAngleFactor -= 0.5;
                sprintf(buffer, "%+5.2f\r\n", kGyroDeltaAngleFactor);
                e_send_uart1_char(buffer, strlen(buffer));
            }
        }
        
        // Read accelerometer.
        tempY = e_get_acc_filtered(1,1);
        accY = (double)(tempY-accYOffset);
        accYOffsetSum += tempY;
        tempZ = e_get_acc_filtered(2,1);
        accZ = (double)(tempZ-accYOffset);
        accAngle = (accY/sqrt((accY*accY) + (accZ*accZ)))*CST_RADIAN;
        accAngle = -accAngle;   // Change the sign to be compatible with the gyro data.
        
        // Read gyroscope.
        getAllAxesGyro(&tempX, &tempY, &tempZ);
        gyroYOffsetSum += tempY;
        gyroDps = rawToDps(tempY-gyroYOffset);
        gyroDeltaAngle = gyroDps*0.005*kGyroDeltaAngleFactor;

        // Compute current inclination angle fusing accelerometer and gyro data.
        // The resulting angle is actually the current error.
        currAngle = (1.0-kGyro)*(accAngle) + kGyro*(prevAngle+gyroDeltaAngle);
        if(abs(currAngle) < angleZeroThr) {
            currAngle = 0.0;
            errorSum = 0.0;
        }
        prevAngle = currAngle;

        prevError = error;
        error = currAngle;

        errorSum += error;
        if(errorSum > iControllerSumMax) {
            errorSum = iControllerSumMax;
        }
        if(errorSum < -iControllerSumMax) {
            errorSum = -iControllerSumMax;
        }

        speed = error*pController;  // P
        speed += ((error - prevError)*dController); // D
        speed += (errorSum*iController); // I
        
        if(speed > maxSpeed) {
            speed = maxSpeed;
        }
        if(speed < -maxSpeed) {
            speed = -maxSpeed;
        }
        
        e_set_speed_left((int)speed);
        e_set_speed_right((int)speed);

        // We can update the offset of the accelerometer and gyroscope continuously by setting the flag "updateOffset".
        // A new offset will be computed every about 5 seconds and it will be filtered.
        offsetCount++;
        if(offsetCount == 1024) {   // 1024 at 200 Hz means about 5 seconds.
            offsetCount = 0;
            if(updateOffset == 1) {
                accYOffset = (double)accYOffset*0.9 + (double)(accYOffsetSum>>10)*0.1;
                gyroYOffset = (double)gyroYOffset*0.9 + (double)(gyroYOffsetSum>>10)*0.1;
            }
            accYOffsetSum = 0;
            gyroYOffsetSum = 0;
        }
        
        LED0 = 0;
        while(getDiffTimeMs() <= 5);   // Wait 5 ms (controller run at 200 Hz).
        resetTime();

        //sprintf(buffer, "%+8.2f;%+8.2f;%+8.2f;%+8.2f;%+8.2f;%+8.2f;%+8.2f;%+d;%+d;%+d\n", getDiffTimeMs(), accAngle, gyroDps, gyroDeltaAngle, errorTot, prevErrorTot, errorTotSum, (int)speed, accYOffset, gyroYOffset);
        //e_send_uart1_char(buffer, strlen(buffer));
        //while(e_uart1_sending());
    }

    return 0;

}

